package dev.fyloz.colorrecipesexplorer.model

import com.fasterxml.jackson.annotation.JsonIgnore
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.CannotDeleteException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import org.springframework.web.multipart.MultipartFile
import javax.persistence.*
import javax.validation.constraints.Min
import javax.validation.constraints.NotBlank
import javax.validation.constraints.Size

const val SIMDUT_FILES_PATH = "pdf/simdut"

@Entity
@Table(name = "material")
data class Material(
    @Id
    @GeneratedValue(strategy = GenerationType.IDENTITY)
    override val id: Long?,

    @Column(unique = true)
    override var name: String,

    @Column(name = "inventory_quantity")
    var inventoryQuantity: Float,

    @Column(name = "mix_type")
    val isMixType: Boolean,

    @ManyToOne
    @JoinColumn(name = "material_type_id")
    var materialType: MaterialType?
) : NamedModel {
    val simdutFilePath
        @JsonIgnore
        @Transient
        get() = "$SIMDUT_FILES_PATH/$name.pdf"
}

open class MaterialSaveDto(
    @field:NotBlank
    val name: String,

    @field:Min(0, message = VALIDATION_SIZE_GE_ZERO)
    val inventoryQuantity: Float,

    val materialTypeId: Long,

    val simdutFile: MultipartFile? = null
) : EntityDto<Material>

open class MaterialUpdateDto(
    val id: Long,

    @field:NotBlank
    val name: String?,

    @field:Min(0, message = VALIDATION_SIZE_GE_ZERO)
    val inventoryQuantity: Float?,

    val materialTypeId: Long?,

    val simdutFile: MultipartFile? = null
) : EntityDto<Material>

data class MaterialOutputDto(
    override val id: Long,
    val name: String,
    val inventoryQuantity: Float,
    val isMixType: Boolean,
    val materialType: MaterialType,
    val simdutUrl: String?
) : Model

data class MaterialQuantityDto(
    val material: Long,

    @field:Min(0, message = VALIDATION_SIZE_GE_ZERO)
    val quantity: Float
)

// === DSL ===

fun material(
    id: Long? = null,
    name: String = "name",
    inventoryQuantity: Float = 0f,
    isMixType: Boolean = false,
    materialType: MaterialType? = materialType(),
    op: Material.() -> Unit = {}
) = Material(id, name, inventoryQuantity, isMixType, materialType).apply(op)

fun material(
    material: Material,
    id: Long? = null,
    name: String? = null,
) = Material(
    id ?: material.id, name
        ?: material.name, material.inventoryQuantity, material.isMixType, material.materialType
)

fun materialSaveDto(
    name: String = "name",
    inventoryQuantity: Float = 0f,
    materialTypeId: Long = 0L,
    simdutFile: MultipartFile? = null,
    op: MaterialSaveDto.() -> Unit = {}
) = MaterialSaveDto(name, inventoryQuantity, materialTypeId, simdutFile).apply(op)

fun materialUpdateDto(
    id: Long = 0L,
    name: String? = "name",
    inventoryQuantity: Float? = 0f,
    materialTypeId: Long? = 0L,
    simdutFile: MultipartFile? = null,
    op: MaterialUpdateDto.() -> Unit = {}
) = MaterialUpdateDto(id, name, inventoryQuantity, materialTypeId, simdutFile).apply(op)

fun materialQuantityDto(
    materialId: Long,
    quantity: Float,
    op: MaterialQuantityDto.() -> Unit = {}
) = MaterialQuantityDto(materialId, quantity).apply(op)

// ==== Exceptions ====
private const
val MATERIAL_NOT_FOUND_EXCEPTION_TITLE = "Material not found"
private const val MATERIAL_ALREADY_EXISTS_EXCEPTION_TITLE = "Material already exists"
private const val MATERIAL_CANNOT_DELETE_EXCEPTION_TITLE = "Cannot delete material"
private const val MATERIAL_EXCEPTION_ERROR_CODE = "material"

fun materialIdNotFoundException(id: Long) =
    NotFoundException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_NOT_FOUND_EXCEPTION_TITLE,
        "A material with the id $id could not be found",
        id
    )

fun materialNameNotFoundException(name: String) =
    NotFoundException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_NOT_FOUND_EXCEPTION_TITLE,
        "A material with the name $name could not be found",
        name,
        "name"
    )

fun materialIdAlreadyExistsException(id: Long) =
    AlreadyExistsException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_ALREADY_EXISTS_EXCEPTION_TITLE,
        "A material with the id $id already exists",
        id
    )

fun materialNameAlreadyExistsException(name: String) =
    AlreadyExistsException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_ALREADY_EXISTS_EXCEPTION_TITLE,
        "A material with the name $name already exists",
        name,
        "name"
    )

fun cannotDeleteMaterialException(material: Material) =
    CannotDeleteException(
        MATERIAL_EXCEPTION_ERROR_CODE,
        MATERIAL_CANNOT_DELETE_EXCEPTION_TITLE,
        "Cannot delete the material ${material.name} because one or more recipes depends on it"
    )
