package dev.fyloz.colorrecipesexplorer.model

import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.CannotDeleteException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.exception.RestException
import org.springframework.http.HttpStatus
import javax.persistence.*

@Entity
@Table(name = "mix_type")
data class MixType(
  @Id
  @GeneratedValue(strategy = GenerationType.IDENTITY)
  override val id: Long?,

  @Column(unique = true)
  override var name: String,

  @OneToOne(cascade = [CascadeType.ALL])
  @JoinColumn(name = "material_id")
  var material: Material
) : NamedModel

// ==== DSL ====
fun mixType(
  id: Long? = null,
  name: String = "name",
  material: Material = material(),
  op: MixType.() -> Unit = {}
) = MixType(id, name, material).apply(op)

fun mixType(
  name: String = "name",
  materialType: MaterialType = materialType(),
  op: MixType.() -> Unit = {}
) = mixType(
  id = null,
  name,
  material = material(name = name, inventoryQuantity = 0f, isMixType = true, materialType = materialType)
).apply(op)

// ==== Exceptions ====
private const val MIX_TYPE_NOT_FOUND_EXCEPTION_TITLE = "Mix type not found"
private const val MIX_TYPE_ALREADY_EXISTS_EXCEPTION_TITLE = "Mix type already exists"
private const val MIX_TYPE_CANNOT_DELETE_EXCEPTION_TITLE = "Cannot delete mix type"
private const val MIX_TYPE_EXCEPTION_ERROR_CODE = "mixtype"

class MixTypeNameAndMaterialTypeNotFoundException(name: String, materialType: MaterialType) :
  RestException(
    "notfound-mixtype-namematerialtype",
    MIX_TYPE_NOT_FOUND_EXCEPTION_TITLE,
    HttpStatus.NOT_FOUND,
    "A mix type with the name $name and material type ${materialType.name} could not be found",
    mapOf(
      "name" to name,
      "materialType" to materialType.name
    )
  )

fun mixTypeIdNotFoundException(id: Long) =
  NotFoundException(
    MIX_TYPE_EXCEPTION_ERROR_CODE,
    MIX_TYPE_NOT_FOUND_EXCEPTION_TITLE,
    "A mix type with the id $id could not be found",
    id
  )

fun mixTypeIdAlreadyExistsException(id: Long) =
  AlreadyExistsException(
    MIX_TYPE_EXCEPTION_ERROR_CODE,
    MIX_TYPE_ALREADY_EXISTS_EXCEPTION_TITLE,
    "A mix type with the id $id already exists",
    id
  )

fun mixTypeNameNotFoundException(name: String) =
  NotFoundException(
    MIX_TYPE_EXCEPTION_ERROR_CODE,
    MIX_TYPE_NOT_FOUND_EXCEPTION_TITLE,
    "A mix type with the name $name could not be found",
    name,
    "name"
  )

fun mixTypeNameAlreadyExistsException(name: String) =
  AlreadyExistsException(
    MIX_TYPE_EXCEPTION_ERROR_CODE,
    MIX_TYPE_ALREADY_EXISTS_EXCEPTION_TITLE,
    "A mix type with the name $name already exists",
    name,
    "name"
  )

fun cannotDeleteMixTypeException(mixType: MixType) =
  CannotDeleteException(
    MIX_TYPE_EXCEPTION_ERROR_CODE,
    MIX_TYPE_CANNOT_DELETE_EXCEPTION_TITLE,
    "Cannot delete the mix type ${mixType.name} because one or more mixes depends on it"
  )
