package dev.fyloz.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.model.account.group
import dev.fyloz.colorrecipesexplorer.repository.RecipeStepRepository
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.TestInstance
import org.junit.jupiter.api.assertThrows
import kotlin.test.assertTrue

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class RecipeStepServiceTest :
  AbstractModelServiceTest<RecipeStep, RecipeStepService, RecipeStepRepository>() {
    override val repository: RecipeStepRepository = mock()
    override val service: RecipeStepService = spy(RecipeStepServiceImpl(repository))

    override val entity: RecipeStep = recipeStep(id = 0L, message = "message")
    override val anotherEntity: RecipeStep = recipeStep(id = 1L, message = "another message")

    // validateGroupInformationSteps()

    @Test
    fun `validateGroupInformationSteps() calls validateSteps() with the given RecipeGroupInformation steps`() {
        withGroupInformation {
            service.validateGroupInformationSteps(this)

            verify(service).validateSteps(this.steps!!)
        }
    }

    @Test
    fun `validateGroupInformationSteps() throws InvalidGroupStepsPositionsException when validateSteps() throws an InvalidStepsPositionsException`() {
        withGroupInformation {
            doAnswer { throw InvalidStepsPositionsException(setOf()) }.whenever(service).validateSteps(this.steps!!)

            assertThrows<InvalidGroupStepsPositionsException> {
                service.validateGroupInformationSteps(this)
            }
        }
    }

    // validateSteps()

    @Test
    fun `validateSteps() throws an InvalidStepsPositionsException when the position of the first step of the given groupInformation is not 1`() {
        assertInvalidStepsPositionsException(
          mutableSetOf(
            recipeStep(id = 0L, position = 0),
            recipeStep(id = 1L, position = 1),
            recipeStep(id = 2L, position = 2),
            recipeStep(id = 3L, position = 3)
          ),
          INVALID_FIRST_STEP_POSITION_ERROR_CODE
        )
    }

    @Test
    fun `validateSteps() throws an InvalidStepsPositionsException when steps positions are duplicated in the given groupInformation`() {
        assertInvalidStepsPositionsException(
          mutableSetOf(
            recipeStep(id = 0L, position = 1),
            recipeStep(id = 1L, position = 2),
            recipeStep(id = 2L, position = 2),
            recipeStep(id = 3L, position = 3)
          ),
          DUPLICATED_STEPS_POSITIONS_ERROR_CODE
        )
    }

    @Test
    fun `validateSteps() throws an InvalidStepsPositionsException when there is a gap between steps positions in the given groupInformation`() {
        assertInvalidStepsPositionsException(
          mutableSetOf(
            recipeStep(id = 0L, position = 1),
            recipeStep(id = 1L, position = 2),
            recipeStep(id = 2L, position = 4),
            recipeStep(id = 3L, position = 5)
          ),
          GAP_BETWEEN_STEPS_POSITIONS_ERROR_CODE
        )
    }

    private fun withGroupInformation(steps: MutableSet<RecipeStep>? = null, test: RecipeGroupInformation.() -> Unit) {
        recipeGroupInformation(
          group = group(id = 0L),
          steps = steps ?: mutableSetOf(
            recipeStep(id = 0L, position = 1),
            recipeStep(id = 1L, position = 2),
            recipeStep(id = 2L, position = 3),
            recipeStep(id = 3L, position = 4)
          )
        ) {
            test()
        }
    }

    private fun assertInvalidStepsPositionsException(steps: MutableSet<RecipeStep>, errorType: String) {
        val exception = assertThrows<InvalidStepsPositionsException> {
            service.validateSteps(steps)
        }

        assertTrue { exception.errors.size == 1 }
        assertTrue { exception.errors.first().type == errorType }
    }
}
